@@ -2,7 +2,7 @@
  *
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2008,2009 Olly Betts
  * Copyright 2006,2007,2008 Lemur Consulting Ltd
  *
  * This program is free software; you can redistribute it and/or
@@ -31,7 +31,6 @@
 #include "serialise-double.h"
 
 #include <xapian/error.h>
-#include <xapian/enquire.h>
 #include <xapian/termiterator.h>
 #include <xapian/version.h>
 #include "vectortermlist.h"
@@ -315,11 +314,13 @@ Xapian::Query::Internal::accumulate_terms(
 			vector<pair<string, Xapian::termpos> > &terms) const
 {
     if (is_leaf(op)) {
-        // We're a leaf, so just return our term.
-        terms.push_back(make_pair(tname, term_pos));
+	// We're a leaf, so just return our term, but skip Query::MatchAllTerms
+	// (which is Query("")).
+	if (!tname.empty())
+	    terms.push_back(make_pair(tname, term_pos));
     } else {
 	subquery_list::const_iterator end = subqs.end();
-        // not a leaf, concatenate results from all subqueries
+	// Not a leaf, concatenate results from all subqueries.
 	for (subquery_list::const_iterator i = subqs.begin(); i != end; ++i) {
 	    (*i)->accumulate_terms(terms);
 	}
@@ -430,11 +431,13 @@ qint_from_vector(Xapian::Query::op op,
     Xapian::Query::Internal * qint = new Xapian::Query::Internal(op, parameter);
     vector<Xapian::Query::Internal *>::const_iterator i;
     for (i = vec.begin(); i != vec.end(); i++) {
-	qint->add_subquery(*i);
-	delete *i;
+	qint->add_subquery_nocopy(*i);
     }
-    qint->end_construction();
-    return qint;
+    Xapian::Query::Internal * r = qint->end_construction();
+    // We're only called during unserialisation, so no simplification should
+    // happen.
+    AssertEq(r, qint);
+    return r;
 }
 
 static Xapian::Query::Internal *
@@ -447,11 +450,13 @@ qint_from_vector(Xapian::Query::op op,
     qint->set_dbl_parameter(dbl_parameter);
     vector<Xapian::Query::Internal *>::const_iterator i;
     for (i = vec.begin(); i != vec.end(); i++) {
-	qint->add_subquery(*i);
-	delete *i;
+	qint->add_subquery_nocopy(*i);
     }
-    qint->end_construction();
-    return qint;
+    Xapian::Query::Internal * r = qint->end_construction();
+    // We're only called during unserialisation, so no simplification should
+    // happen.
+    AssertEq(r, qint);
+    return r;
 }
 
 Xapian::Query::Internal *
@@ -569,25 +574,6 @@ Xapian::Query::Internal::unserialise(const string &)
 }
 #endif
 
-/** Swap the contents of this with another Xapian::Query::Internal,
- *  in a way which is guaranteed not to throw.  This is
- *  used with the assignment operator to make it exception
- *  safe.
- *  It's important to adjust swap with any addition of
- *  member variables!
- */
-void
-Xapian::Query::Internal::swap(Xapian::Query::Internal &other)
-{
-    std::swap(op, other.op);
-    subqs.swap(other.subqs);
-    std::swap(parameter, other.parameter);
-    std::swap(tname, other.tname);
-    std::swap(str_parameter, other.str_parameter);
-    std::swap(term_pos, other.term_pos);
-    std::swap(wqf, other.wqf);
-}
-
 Xapian::Query::Internal::Internal(const Xapian::Query::Internal &copyme)
 	: Xapian::Internal::RefCntBase(),
 	  op(copyme.op),
@@ -776,7 +762,7 @@ Xapian::Query::Internal::simplify_matchnothing()
 Xapian::Query::Internal *
 Xapian::Query::Internal::simplify_query()
 {
-    DEBUGCALL(API, bool, "Xapian::Query::Internal::simplify_query", "");
+    DEBUGCALL(API, Xapian::Query::Internal *, "Xapian::Query::Internal::simplify_query", "");
 
     // Simplify any MatchNothing nodes.
     if (simplify_matchnothing()) {
@@ -801,11 +787,15 @@ Xapian::Query::Internal::simplify_query()
 	    // so we leave it to be removed.
 	    break;
 	case OP_PHRASE: case OP_NEAR:
+	    // Empty and single subquery OP_PHRASE and OP_NEAR get handled
+	    // below.
+	    if (subqs.size() <= 1) break;
+
 	    // Default to the number of subqueries.
 	    if (!parameter) parameter = subqs.size();
-	    // Flatten out sub queries if this is a phrase (or near) operation.
-	    flatten_subqs();
-	    break;
+
+	    // Flatten out sub queries of a phrase or near operation.
+	    return flatten_subqs();
 	case OP_ELITE_SET:
 	    if (!parameter) {
 		// Default to sqrt(number of subqueries), or a minimum of 10.
@@ -892,43 +882,62 @@ Xapian::Query::Internal::collapse_subqs()
 }
 
 /// Change, eg, A NEAR (B AND C) to (A NEAR B) AND (A NEAR C)
-void
+Xapian::Query::Internal *
 Xapian::Query::Internal::flatten_subqs()
 {
     Assert(op == Xapian::Query::OP_NEAR || op == Xapian::Query::OP_PHRASE);
 
     subquery_list::iterator sq;
-    for (sq = subqs.begin(); sq != subqs.end(); sq++) {
+    for (sq = subqs.begin(); sq != subqs.end(); ++sq) {
 	if (!is_leaf((*sq)->op)) break;
     }
 
-    if (sq != subqs.end()) {
-	if ((*sq)->op == Xapian::Query::OP_NEAR ||
-	    (*sq)->op == Xapian::Query::OP_PHRASE) {
-	    // FIXME: A PHRASE (B PHRASE C) -> (A PHRASE B) AND (B PHRASE C)?
-	    throw Xapian::UnimplementedError("Can't use NEAR/PHRASE with a subexpression containing NEAR or PHRASE");
-	}
+    if (sq == subqs.end()) return this;
 
-	AutoPtr<Xapian::Query::Internal> flattenme(*sq);
-	*sq = 0;
-
-	// New query to build up.
-	Xapian::Query::Internal newq(flattenme->op, 0);
-
-	subquery_list::iterator j;
-	for (j = flattenme->subqs.begin(); j != flattenme->subqs.end(); ++j) {
-	    *sq = *j;
-	    *j = 0;
-	    flatten_subqs();
-	    newq.add_subquery(this);
-	    delete *sq;
-	    *sq = 0;
-	}
-
-	Xapian::Query::Internal * newq2 = newq.end_construction();
-	Assert(newq2);
-	this->swap(*newq2);
+    if ((*sq)->op == Xapian::Query::OP_NEAR ||
+	(*sq)->op == Xapian::Query::OP_PHRASE) {
+	// FIXME: A PHRASE (B PHRASE C) -> (A PHRASE B) AND (B PHRASE C)?
+	throw Xapian::UnimplementedError("Can't use NEAR/PHRASE with a subexpression containing NEAR or PHRASE");
     }
+
+    AutoPtr<Xapian::Query::Internal> flattenme(*sq);
+    *sq = 0;
+
+    subquery_list::iterator j;
+    for (j = flattenme->subqs.begin(); j != flattenme->subqs.end(); ++j) {
+	*sq = *j;
+	*j = 0;
+	AutoPtr<Xapian::Query::Internal> newq(new Xapian::Query::Internal(*this));
+	delete *sq;
+	*sq = 0;
+	newq = newq->flatten_subqs();
+	*j = newq.release();
+    }
+
+    if (flattenme->op == OP_AND ||
+	flattenme->op == OP_OR ||
+	flattenme->op == OP_XOR) {
+	size_t i = flattenme->subqs.size();
+	do {
+	    --i;
+	    Xapian::Query::Internal * q = flattenme->subqs[i];
+	    if (flattenme->op == q->op) {
+		subquery_list::iterator k;
+		for (k = q->subqs.begin(), ++k;
+		     k != q->subqs.end();
+		     ++k) {
+		    flattenme->subqs.push_back(0);
+		    flattenme->subqs.back() = *k;
+		    *k = 0;
+		}
+		flattenme->subqs[i] = q->subqs[0];
+		q->subqs.clear();
+		delete q;
+	    }
+	} while (i != 0);
+    }
+
+    return flattenme.release();
 }
 
 void
@@ -948,6 +957,24 @@ Xapian::Query::Internal::add_subquery(const Xapian::Query::Internal * subq)
     }
 }
 
+void
+Xapian::Query::Internal::add_subquery_nocopy(Xapian::Query::Internal * subq)
+{
+    Assert(!is_leaf(op));
+    if (subq == 0) {
+	subqs.push_back(0);
+    } else if (op == subq->op && (op == OP_AND || op == OP_OR || op == OP_XOR)) {
+	// Distribute the subquery.
+	for (subquery_list::const_iterator i = subq->subqs.begin();
+	     i != subq->subqs.end(); i++) {
+	    add_subquery(*i);
+	}
+	delete subq;
+    } else {
+	subqs.push_back(subq);
+    }
+}
+
 void
 Xapian::Query::Internal::set_dbl_parameter(double dbl_parameter_)
 {
