@@ -200,7 +200,7 @@ FlintLock::lock(bool exclusive) {
     close(fds[0]);
 
     int status;
-    while (waitpid(pid, &status, 0) < 0) {
+    while (waitpid(child, &status, 0) < 0) {
 	if (errno != EINTR) break;
     }
 
