@@ -2,7 +2,7 @@
  *
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2007 Olly Betts
+ * Copyright 2002,2003,2004,2005,2007,2009 Olly Betts
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -47,7 +47,7 @@ class FlintRecordTable : public FlintTable {
 	 *  @param readonly_      - whether to open the table for read only
 	 *                          access.
 	 */
-	FlintRecordTable(string path_, bool readonly_)
+	FlintRecordTable(const string & path_, bool readonly_)
 	    : FlintTable(path_ + "/record.", readonly_, Z_DEFAULT_STRATEGY) { }
 
 	/** Retrieve a document from the table.
