@@ -2,7 +2,7 @@
  *
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2009 Olly Betts
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -77,11 +77,8 @@ PWRITE_PROTOTYPE
 
 #include <algorithm>  // for std::min()
 #include <string>
-#include <vector>
 
-using std::min;
-using std::string;
-using std::vector;
+using namespace std;
 
 // Only try to compress tags longer than this many bytes.
 const size_t COMPRESS_MIN = 4;
@@ -250,7 +247,7 @@ FlintTable::read_block(uint4 n, byte * p) const
 	    /* Read part of the block, which is not an error.  We should
 	     * continue reading the rest of the block.
 	     */
-	    m -= bytes_read;
+	    m -= int(bytes_read);
 	    p += bytes_read;
 	    offset += bytes_read;
 	}
@@ -288,7 +285,7 @@ FlintTable::write_block(uint4 n, const byte * p) const
 
     if (both_bases) {
 	// Delete the old base before modifying the database.
-	sys_unlink(name + "base" + other_base_letter);
+	sys_unlink(name + "base" + other_base_letter());
 	both_bases = false;
 	latest_revision_number = revision_number;
     }
@@ -313,7 +310,7 @@ FlintTable::write_block(uint4 n, const byte * p) const
 	    /* Wrote part of the block, which is not an error.  We should
 	     * continue writing the rest of the block.
 	     */
-	    m -= bytes_written;
+	    m -= int(bytes_written);
 	    p += bytes_written;
 	    offset += bytes_written;
 	}
@@ -1024,7 +1021,7 @@ FlintTable::add(const string &key, string tag, bool already_compressed)
     DEBUGCALL(DB, bool, "FlintTable::add", key << ", " << tag);
     Assert(writable);
 
-    if (handle == -1) create_and_open(block_size);
+    if (handle < 0) create_and_open(block_size);
 
     form_key(key);
 
@@ -1039,56 +1036,27 @@ FlintTable::add(const string &key, string tag, bool already_compressed)
 	CompileTimeAssert(DONT_COMPRESS != Z_RLE);
 #endif
 
-	z_stream stream;
+	lazy_alloc_deflate_zstream();
 
-	stream.zalloc = reinterpret_cast<alloc_func>(0);
-	stream.zfree = reinterpret_cast<free_func>(0);
-	stream.opaque = (voidpf)0;
-
-	// -15 means raw deflate with 32K LZ77 window (largest)
-	// memLevel 9 is the highest (8 is default)
-	int err;
-	err = deflateInit2(&stream, Z_DEFAULT_COMPRESSION, Z_DEFLATED, -15, 9,
-			   compress_strategy);
-	if (err != Z_OK) {
-	    if (err == Z_MEM_ERROR) throw std::bad_alloc();
-	    string msg = "deflateInit2 failed";
-	    if (stream.msg) {
-		msg += " (";
-		msg += stream.msg;
-		msg += ')';
-	    }
-	    throw Xapian::DatabaseError(msg);
-	}
-
-	stream.next_in = (Bytef *)const_cast<char *>(tag.data());
-	stream.avail_in = (uInt)tag.size();
+	// zlib takes a non-const pointer to the input, but doesn't modify it.
+	char * non_const_tag = const_cast<char *>(tag.data());
+	deflate_zstream->next_in = reinterpret_cast<Byte *>(non_const_tag);
+	deflate_zstream->avail_in = uInt(tag.size());
 
 	// If compressed size is >= tag.size(), we don't want to compress.
 	unsigned long blk_len = tag.size() - 1;
 	unsigned char * blk = new unsigned char[blk_len];
-	stream.next_out = blk;
-	stream.avail_out = (uInt)blk_len;
+	deflate_zstream->next_out = blk;
+	deflate_zstream->avail_out = uInt(blk_len);
 
-	err = deflate(&stream, Z_FINISH);
+	int err = deflate(deflate_zstream, Z_FINISH);
 	if (err == Z_STREAM_END) {
 	    // If deflate succeeded, then the output was at least one byte
 	    // smaller than the input.
-	    tag.assign(reinterpret_cast<const char *>(blk), stream.total_out);
+	    tag.assign(reinterpret_cast<const char *>(blk), deflate_zstream->total_out);
 	    compressed = true;
-	    err = deflateEnd(&stream);
-	    if (err != Z_OK) {
-		string msg = "deflateEnd failed";
-		if (stream.msg) {
-		    msg += " (";
-		    msg += stream.msg;
-		    msg += ')';
-		}
-		throw Xapian::DatabaseError(msg);
-	    }
 	} else {
 	    // Deflate failed - presumably the data wasn't compressible.
-	    (void)deflateEnd(&stream);
 	}
 
 	delete [] blk;
@@ -1173,7 +1141,7 @@ FlintTable::del(const string &key)
     DEBUGCALL(DB, bool, "FlintTable::del", key);
     Assert(writable);
 
-    if (handle == -1) RETURN(false);
+    if (handle < 0) RETURN(false);
 
     // We can't delete a key which we is too long for us to store.
     if (key.size() > FLINT_BTREE_MAX_KEY_LEN) RETURN(false);
@@ -1200,6 +1168,8 @@ FlintTable::get_exact_entry(const string &key, string & tag) const
     DEBUGCALL(DB, bool, "FlintTable::get_exact_entry", key << ", " << tag);
     Assert(!key.empty());
 
+    if (handle < 0) RETURN(false);
+
     // An oversized key can't exist, so attempting to search for it should fail.
     if (key.size() > FLINT_BTREE_MAX_KEY_LEN) RETURN(false);
 
@@ -1271,71 +1241,51 @@ FlintTable::read_tag(Cursor_ * C_, string *tag, bool keep_compressed) const
 
     Bytef buf[8192];
 
-    z_stream stream;
-    stream.next_out = buf;
-    stream.avail_out = (uInt)sizeof(buf);
+    lazy_alloc_inflate_zstream();
 
-    stream.zalloc = reinterpret_cast<alloc_func>(0);
-    stream.zfree = reinterpret_cast<free_func>(0);
-
-    stream.next_in = Z_NULL;
-    stream.avail_in = 0;
-
-    int err = inflateInit2(&stream, -15);
-    if (err != Z_OK) {
-	if (err == Z_MEM_ERROR) throw std::bad_alloc();
-	string msg = "inflateInit2 failed";
-	if (stream.msg) {
-	    msg += " (";
-	    msg += stream.msg;
-	    msg += ')';
-	}
-	throw Xapian::DatabaseError(msg);
-    }
-
-    stream.next_in = (Bytef*)const_cast<char *>(tag->data());
-    stream.avail_in = (uInt)tag->size();
+    // zlib takes a non-const pointer to the input, but doesn't modify it.
+    char * non_const_tag = const_cast<char *>(tag->data());
+    inflate_zstream->next_in = reinterpret_cast<Byte *>(non_const_tag);
+    inflate_zstream->avail_in = uInt(tag->size());
 
+    int err = Z_OK;
     while (err != Z_STREAM_END) {
-	stream.next_out = buf;
-	stream.avail_out = (uInt)sizeof(buf);
-	err = inflate(&stream, Z_SYNC_FLUSH);
-	if (err == Z_BUF_ERROR && stream.avail_in == 0) {
-	    DEBUGLINE(DB, "Z_BUF_ERROR - faking checksum of " << stream.adler);
+	inflate_zstream->next_out = buf;
+	inflate_zstream->avail_out = uInt(sizeof(buf));
+	err = inflate(inflate_zstream, Z_SYNC_FLUSH);
+	if (err == Z_BUF_ERROR && inflate_zstream->avail_in == 0) {
+	    DEBUGLINE(DB, "Z_BUF_ERROR - faking checksum of " << inflate_zstream->adler);
 	    Bytef header2[4];
-	    setint4(header2, 0, stream.adler);
-	    stream.next_in = header2;
-	    stream.avail_in = 4;
-	    err = inflate(&stream, Z_SYNC_FLUSH);
+	    setint4(header2, 0, inflate_zstream->adler);
+	    inflate_zstream->next_in = header2;
+	    inflate_zstream->avail_in = 4;
+	    err = inflate(inflate_zstream, Z_SYNC_FLUSH);
 	    if (err == Z_STREAM_END) break;
 	}
 
 	if (err != Z_OK && err != Z_STREAM_END) {
 	    if (err == Z_MEM_ERROR) throw std::bad_alloc();
 	    string msg = "inflate failed";
-	    if (stream.msg) {
+	    if (inflate_zstream->msg) {
 		msg += " (";
-		msg += stream.msg;
+		msg += inflate_zstream->msg;
 		msg += ')';
 	    }
 	    throw Xapian::DatabaseError(msg);
 	}
 
 	utag.append(reinterpret_cast<const char *>(buf),
-		    stream.next_out - buf);
+		    inflate_zstream->next_out - buf);
     }
-    if (utag.size() != stream.total_out) {
+    if (utag.size() != inflate_zstream->total_out) {
 	string msg = "compressed tag didn't expand to the expected size: ";
 	msg += om_tostring(utag.size());
 	msg += " != ";
 	// OpenBSD's zlib.h uses off_t instead of uLong for total_out.
-	msg += om_tostring((size_t)stream.total_out);
+	msg += om_tostring(size_t(inflate_zstream->total_out));
 	throw Xapian::DatabaseCorruptError(msg);
     }
 
-    err = inflateEnd(&stream);
-    if (err != Z_OK) abort();
-
     swap(*tag, utag);
 
     return false;
@@ -1351,7 +1301,7 @@ FlintTable::set_full_compaction(bool parity)
 }
 
 FlintCursor * FlintTable::cursor_get() const {
-    if (handle == -1) return NULL;
+    if (handle < 0) return NULL;
     // FIXME Ick - casting away const is nasty
     return new FlintCursor(const_cast<FlintTable *>(this));
 }
@@ -1361,7 +1311,7 @@ FlintCursor * FlintTable::cursor_get() const {
 bool
 FlintTable::basic_open(bool revision_supplied, flint_revision_number_t revision_)
 {
-    int ch = 'X'; /* will be 'A' or 'B' */
+    char ch = 'X'; /* will be 'A' or 'B' */
 
     {
 	const size_t BTREE_BASES = 2;
@@ -1386,7 +1336,7 @@ FlintTable::basic_open(bool revision_supplied, flint_revision_number_t revision_
 	}
 
 	if (!valid_base) {
-	    if (handle != -1) {
+	    if (handle >= 0) {
 		::close(handle);
 		handle = -1;
 	    }
@@ -1490,7 +1440,7 @@ FlintTable::read_root()
 	byte * p = C[0].p;
 	Assert(p);
 
-	/* clear block - shouldn't be neccessary, but is a bit nicer,
+	/* clear block - shouldn't be necessary, but is a bit nicer,
 	 * and means that the same operations should always produce
 	 * the same database. */
 	memset(p, 0, block_size);
@@ -1530,6 +1480,9 @@ FlintTable::do_open_to_write(bool revision_supplied,
 			     flint_revision_number_t revision_,
 			     bool create_db)
 {
+    if (handle == -2) {
+	throw Xapian::DatabaseError("Database has been closed");
+    }
     int flags = O_RDWR | O_BINARY;
     if (create_db) flags |= O_CREAT | O_TRUNC;
     handle = ::open((name + "DB").c_str(), flags, 0666);
@@ -1576,9 +1529,6 @@ FlintTable::do_open_to_write(bool revision_supplied,
 
     buffer = zeroed_new(block_size);
 
-    // swap for writing
-    other_base_letter = base_letter == 'A' ? 'B' : 'A';
-
     changed_n = 0;
     changed_c = DIR_START;
     seq_count = SEQ_START_POINT;
@@ -1586,7 +1536,7 @@ FlintTable::do_open_to_write(bool revision_supplied,
     return true;
 }
 
-FlintTable::FlintTable(string path_, bool readonly_,
+FlintTable::FlintTable(const string & path_, bool readonly_,
 		       int compress_strategy_, bool lazy_)
 	: revision_number(0),
 	  item_count(0),
@@ -1602,7 +1552,6 @@ FlintTable::FlintTable(string path_, bool readonly_,
 	  kt(0),
 	  buffer(0),
 	  base(),
-	  other_base_letter(0),
 	  name(path_),
 	  seq_count(0),
 	  changed_n(0),
@@ -1613,9 +1562,87 @@ FlintTable::FlintTable(string path_, bool readonly_,
 	  writable(!readonly_),
 	  split_p(0),
 	  compress_strategy(compress_strategy_),
+	  deflate_zstream(NULL),
+	  inflate_zstream(NULL),
 	  lazy(lazy_)
 {
-    DEBUGCALL(DB, void, "FlintTable::Btree", path_ << ", " << readonly_);
+    DEBUGCALL(DB, void, "FlintTable::FlintTable",
+	      path_ << ", " << readonly_ << ", " <<
+	      compress_strategy_ << ", " << lazy_);
+}
+
+void
+FlintTable::lazy_alloc_deflate_zstream() const {
+    if (usual(deflate_zstream)) {
+	if (usual(deflateReset(deflate_zstream) == Z_OK)) return;
+	// Try to recover by deleting the stream and starting from scratch.
+	delete deflate_zstream;
+    }
+
+    deflate_zstream = new z_stream;
+
+    deflate_zstream->zalloc = reinterpret_cast<alloc_func>(0);
+    deflate_zstream->zfree = reinterpret_cast<free_func>(0);
+    deflate_zstream->opaque = voidpf(0);
+
+    // -15 means raw deflate with 32K LZ77 window (largest)
+    // memLevel 9 is the highest (8 is default)
+    int err;
+    err = deflateInit2(deflate_zstream, Z_DEFAULT_COMPRESSION, Z_DEFLATED,
+		       -15, 9, compress_strategy);
+    if (rare(err != Z_OK)) {
+	if (err == Z_MEM_ERROR) {
+	    delete deflate_zstream;
+	    deflate_zstream = 0;
+	    throw std::bad_alloc();
+	}
+	string msg = "deflateInit2 failed (";
+	if (deflate_zstream->msg) {
+	    msg += deflate_zstream->msg;
+	} else {
+	    msg += om_tostring(err);
+	}
+	msg += ')';
+	delete deflate_zstream;
+	deflate_zstream = 0;
+	throw Xapian::DatabaseError(msg);
+    }
+}
+
+void
+FlintTable::lazy_alloc_inflate_zstream() const {
+    if (usual(inflate_zstream)) {
+	if (usual(inflateReset(inflate_zstream) == Z_OK)) return;
+	// Try to recover by deleting the stream and starting from scratch.
+	delete inflate_zstream;
+    }
+
+    inflate_zstream = new z_stream;
+
+    inflate_zstream->zalloc = reinterpret_cast<alloc_func>(0);
+    inflate_zstream->zfree = reinterpret_cast<free_func>(0);
+
+    inflate_zstream->next_in = Z_NULL;
+    inflate_zstream->avail_in = 0;
+
+    int err = inflateInit2(inflate_zstream, -15);
+    if (rare(err != Z_OK)) {
+	if (err == Z_MEM_ERROR) {
+	    delete inflate_zstream;
+	    inflate_zstream = 0;
+	    throw std::bad_alloc();
+	}
+	string msg = "inflateInit2 failed (";
+	if (inflate_zstream->msg) {
+	    msg += inflate_zstream->msg;
+	} else {
+	    msg += om_tostring(err);
+	}
+	msg += ')';
+	delete inflate_zstream;
+	inflate_zstream = 0;
+	throw Xapian::DatabaseError(msg);
+    }
 }
 
 bool
@@ -1665,6 +1692,9 @@ void
 FlintTable::create_and_open(unsigned int block_size_)
 {
     DEBUGCALL(DB, void, "FlintTable::create_and_open", block_size_);
+    if (handle == -2) {
+	throw Xapian::DatabaseError("Database has been closed");
+    }
     Assert(writable);
     close();
 
@@ -1695,18 +1725,36 @@ FlintTable::create_and_open(unsigned int block_size_)
 FlintTable::~FlintTable() {
     DEBUGCALL(DB, void, "FlintTable::~FlintTable", "");
     FlintTable::close();
+
+    if (deflate_zstream) {
+	// Errors which we care about have already been handled, so just ignore
+	// any which get returned here.
+	(void) deflateEnd(deflate_zstream);
+	delete deflate_zstream;
+    }
+
+    if (inflate_zstream) {
+	// Errors which we care about have already been handled, so just ignore
+	// any which get returned here.
+	(void) inflateEnd(inflate_zstream);
+	delete inflate_zstream;
+    }
 }
 
-void FlintTable::close() {
+void FlintTable::close(bool permanent) {
     DEBUGCALL(DB, void, "FlintTable::close", "");
 
-    if (handle != -1) {
+    if (handle >= 0) {
 	// If an error occurs here, we just ignore it, since we're just
 	// trying to free everything.
 	(void)::close(handle);
 	handle = -1;
     }
 
+    if (permanent) {
+	handle = -2;
+    }
+
     for (int j = level; j >= 0; j--) {
 	delete [] C[j].p;
 	C[j].p = 0;
@@ -1730,7 +1778,7 @@ FlintTable::commit(flint_revision_number_t revision)
 	throw Xapian::DatabaseError("New revision too low");
     }
 
-    if (handle == -1) {
+    if (handle < 0) {
 	latest_revision_number = revision_number = revision;
 	return;
     }
@@ -1747,79 +1795,81 @@ FlintTable::commit(flint_revision_number_t revision)
 	}
     }
 
-    if (!flint_io_sync(handle)) {
-	(void)::close(handle);
-	handle = -1;
-	throw Xapian::DatabaseError("Can't commit new revision - failed to flush DB to disk");
-    }
-
     if (Btree_modified) {
 	faked_root_block = false;
     }
 
-    if (faked_root_block) {
-	/* We will use a dummy bitmap. */
-	base.clear_bit_map();
-    }
-
-    base.set_revision(revision);
-    base.set_root(C[level].n);
-    base.set_level(level);
-    base.set_item_count(item_count);
-    base.set_have_fakeroot(faked_root_block);
-    base.set_sequential(sequential);
-
-    {
-	int tmp = base_letter;
-	base_letter = other_base_letter;
-	other_base_letter = tmp;
-    }
-    both_bases = true;
-    latest_revision_number = revision_number = revision;
-    root = C[level].n;
-
-    Btree_modified = false;
-
-    for (int i = 0; i < BTREE_CURSOR_LEVELS; ++i) {
-	C[i].n = BLK_UNUSED;
-	C[i].c = -1;
-	C[i].rewrite = false;
-    }
-
-    // Save to "<table>.tmp" and then rename to "<table>.base<letter>" so that
-    // a reader can't try to read a partially written base file.
-    string tmp = name;
-    tmp += "tmp";
-    string basefile = name;
-    basefile += "base";
-    basefile += char(base_letter);
-    base.write_to_file(tmp);
-#if defined __WIN32__
-    if (msvc_posix_rename(tmp.c_str(), basefile.c_str()) < 0) {
-#else
-    if (rename(tmp.c_str(), basefile.c_str()) < 0) {
-#endif
-	// With NFS, rename() failing may just mean that the server crashed
-	// after successfully renaming, but before reporting this, and then
-	// the retried operation fails.  So we need to check if the source
-	// file still exists, which we do by calling unlink(), since we want
-	// to remove the temporary file anyway.
-	int saved_errno = errno;
-	if (unlink(tmp) == 0 || errno != ENOENT) {
-	    string msg("Couldn't update base file ");
-	    msg += basefile;
-	    msg += ": ";
-	    msg += strerror(saved_errno);
-	    throw Xapian::DatabaseError(msg);
+    try {
+	if (faked_root_block) {
+	    /* We will use a dummy bitmap. */
+	    base.clear_bit_map();
 	}
+
+	base.set_revision(revision);
+	base.set_root(C[level].n);
+	base.set_level(level);
+	base.set_item_count(item_count);
+	base.set_have_fakeroot(faked_root_block);
+	base.set_sequential(sequential);
+
+	base_letter = other_base_letter();
+
+	both_bases = true;
+	latest_revision_number = revision_number = revision;
+	root = C[level].n;
+
+	Btree_modified = false;
+
+	for (int i = 0; i < BTREE_CURSOR_LEVELS; ++i) {
+	    C[i].n = BLK_UNUSED;
+	    C[i].c = -1;
+	    C[i].rewrite = false;
+	}
+
+	if (!flint_io_sync(handle)) {
+	    (void)::close(handle);
+	    handle = -1;
+	    throw Xapian::DatabaseError("Can't commit new revision - failed to flush DB to disk");
+	}
+
+	// Save to "<table>.tmp" and then rename to "<table>.base<letter>" so
+	// that a reader can't try to read a partially written base file.
+	string tmp = name;
+	tmp += "tmp";
+	string basefile = name;
+	basefile += "base";
+	basefile += char(base_letter);
+	base.write_to_file(tmp);
+#if defined __WIN32__
+	if (msvc_posix_rename(tmp.c_str(), basefile.c_str()) < 0) {
+#else
+	if (rename(tmp.c_str(), basefile.c_str()) < 0) {
+#endif
+	    // With NFS, rename() failing may just mean that the server crashed
+	    // after successfully renaming, but before reporting this, and then
+	    // the retried operation fails.  So we need to check if the source
+	    // file still exists, which we do by calling unlink(), since we want
+	    // to remove the temporary file anyway.
+	    int saved_errno = errno;
+	    if (unlink(tmp) == 0 || errno != ENOENT) {
+		string msg("Couldn't update base file ");
+		msg += basefile;
+		msg += ": ";
+		msg += strerror(saved_errno);
+		throw Xapian::DatabaseError(msg);
+	    }
+	}
+	base.commit();
+
+	read_root();
+
+	changed_n = 0;
+	changed_c = DIR_START;
+	seq_count = SEQ_START_POINT;
+    } catch(...) {
+	FlintTable::close();
+	throw;
     }
-    base.commit();
-
-    read_root();
-
-    changed_n = 0;
-    changed_c = DIR_START;
-    seq_count = SEQ_START_POINT;
 }
 
 void
@@ -1828,7 +1878,7 @@ FlintTable::cancel()
     DEBUGCALL(DB, void, "FlintTable::cancel", "");
     Assert(writable);
 
-    if (handle == -1) {
+    if (handle < 0) {
 	latest_revision_number = revision_number; // FIXME: we can end up reusing a revision if we opened a btree at an older revision, start to modify it, then cancel...
 	return;
     }
@@ -1837,7 +1887,7 @@ FlintTable::cancel()
 
     string err_msg;
     if (!base.read(name, base_letter, err_msg)) {
-	throw Xapian::DatabaseCorruptError("Couldn't reread base " + base_letter);
+	throw Xapian::DatabaseCorruptError(string("Couldn't reread base ") + base_letter);
     }
 
     revision_number =  base.get_revision();
@@ -1867,6 +1917,9 @@ FlintTable::cancel()
 bool
 FlintTable::do_open_to_read(bool revision_supplied, flint_revision_number_t revision_)
 {
+    if (handle == -2) {
+	throw Xapian::DatabaseError("Database has been closed");
+    }
     handle = ::open((name + "DB").c_str(), O_RDONLY | O_BINARY);
     if (handle < 0) {
 	if (lazy) {
@@ -1986,7 +2039,8 @@ FlintTable::prev_for_sequential(Cursor_ * C_, int /*dummy*/) const
 	    } else {
 		read_block(n, p);
 	    }
-	    if (REVISION(p) > 1) {
+	    if (writable) AssertEq(revision_number, latest_revision_number);
+	    if (REVISION(p) > revision_number + writable) {
 		set_overwritten();
 		return false;
 	    }
@@ -2038,7 +2092,9 @@ FlintTable::next_for_sequential(Cursor_ * C_, int /*dummy*/) const
 	    } else {
 		read_block(n, p);
 	    }
-	    if (REVISION(p) > 1) {
+
+	    if (writable) AssertEq(revision_number, latest_revision_number);
+	    if (REVISION(p) > revision_number + writable) {
 		set_overwritten();
 		return false;
 	    }
