@@ -164,6 +164,12 @@ BitReader::decode_interpolative(vector<Xapian::termpos> & pos, int j, int k)
     }
 }
 
+static inline bool
+is_user_metadata_key(const string & key)
+{
+    return key.size() > 1 && key[0] == '\0' && key[1] != '\xff';
+}
+
 int
 main(int argc, char **argv)
 {
@@ -277,7 +283,7 @@ static size_t
 check_table(string filename, int opts)
 {
     size_t p = filename.find_last_of('/');
-#ifdef __WIN32__
+#if defined __WIN32__ || defined __EMX__
     if (p == string::npos) p = 0;
     p = filename.find_last_of('\\', p);
 #endif
@@ -308,37 +314,50 @@ check_table(string filename, int opts)
 	Xapian::docid lastdid = 0;
 	Xapian::termcount termfreq = 0, collfreq = 0;
 	Xapian::termcount tf = 0, cf = 0;
+	bool have_metainfo_key = false;
 
-	// The first key/tag pair should be the METAINFO.
+	// The first key/tag pair should be the METAINFO - though this may be
+	// missing if the table only contains user-metadata.
 	if (!cursor->after_end()) {
-	    if (cursor->current_key != string("", 1)) {
-		cout << "METAINFO key missing from postlist table" << endl;
-		return errors + 1;
+	    if (cursor->current_key == string("", 1)) {
+		have_metainfo_key = true;
+		cursor->read_tag();
+		// Check format of the METAINFO key.
+		Xapian::docid did;
+		flint_totlen_t totlen;
+		const char * data = cursor->current_tag.data();
+		const char * end = data + cursor->current_tag.size();
+		if (!unpack_uint(&data, end, &did)) {
+		    cout << "Tag containing meta information is corrupt." << endl;
+		    return errors + 1;
+		}
+		if (!unpack_uint_last(&data, end, &totlen)) {
+		    cout << "Tag containing meta information is corrupt." << endl;
+		    return errors + 1;
+		}
+		if (data != end) {
+		    cout << "Tag containing meta information is corrupt." << endl;
+		    return errors + 1;
+		}
+		cursor->next();
 	    }
-	    cursor->read_tag();
-	    // Check format of the METAINFO key.
-	    Xapian::docid did;
-	    flint_totlen_t totlen;
-	    const char * data = cursor->current_tag.data();
-	    const char * end = data + cursor->current_tag.size();
-	    if (!unpack_uint(&data, end, &did)) {
-		cout << "Tag containing meta information is corrupt." << endl;
-		return errors + 1;
-	    }
-	    if (!unpack_uint_last(&data, end, &totlen)) {
-		cout << "Tag containing meta information is corrupt." << endl;
-		return errors + 1;
-	    }
-	    if (data != end) {
-		cout << "Tag containing meta information is corrupt." << endl;
-		return errors + 1;
-	    }
-	    cursor->next();
 	}
 
 	while (!cursor->after_end()) {
 	    string & key = cursor->current_key;
 
+	    if (is_user_metadata_key(key)) {
+		// User metadata can be anything, so we can't do any particular
+		// checks on it.
+		cursor->next();
+		continue;
+	    }
+
+	    if (!have_metainfo_key) {
+		cout << "METAINFO key missing from postlist table" << endl;
+		return errors + 1;
+	    }
+
 	    const char * pos, * end;
 
 	    // Get term from key.
@@ -757,14 +776,14 @@ check_table(string filename, int opts)
 		vector<Xapian::termpos>::const_iterator current_pos = positions.begin();
 		Xapian::termpos lastpos = *current_pos++;
 		while (current_pos != positions.end()) {
-		    Xapian::termpos pos = *current_pos++;
-		    if (pos <= lastpos) {
+		    Xapian::termpos termpos = *current_pos++;
+		    if (termpos <= lastpos) {
 			cout << tablename << " table: Positions not strictly monotonically increasing" << endl;
 			++errors;
 			cursor->next();
 			continue;
 		    }
-		    lastpos = pos;
+		    lastpos = termpos;
 		}
 	    }
 
