@@ -161,16 +161,20 @@ merge_postlists(FlintTable * out, vector<Xapian::docid>::const_iterator offset,
     for ( ; b != e; ++b, ++offset) {
 	FlintTable *in = new FlintTable(*b, true);
 	in->open();
-	if (in->get_entry_count()) {
-	    // PostlistCursor takes ownership of FlintTable in and is
-	    // responsible for deleting it.
-	    PostlistCursor * cur = new PostlistCursor(in, *offset);
-	    // Merge the METAINFO tags from each database into one.
-	    // They have a key consisting of a single zero byte, which will
-	    // always be the first key.
-	    if (!is_metainfo_key(cur->key)) {
-		throw Xapian::DatabaseCorruptError("No METAINFO item in postlist table.");
-	    }
+	if (!in->get_entry_count()) {
+	    // Skip empty tables.
+	    delete in;
+	    continue;
+	}
+
+	// PostlistCursor takes ownership of FlintTable in and is
+	// responsible for deleting it.
+	PostlistCursor * cur = new PostlistCursor(in, *offset);
+	// Merge the METAINFO tags from each database into one.
+	// They have a key consisting of a single zero byte.
+	// They may be absent, if the database contains no documents.  If it
+	// has user metadata we'll still get here.
+	if (is_metainfo_key(cur->key)) {
 	    const char * data = cur->tag.data();
 	    const char * end = data + cur->tag.size();
 	    Xapian::docid dummy_did = 0;
@@ -182,16 +186,14 @@ merge_postlists(FlintTable * out, vector<Xapian::docid>::const_iterator offset,
 		throw Xapian::DatabaseCorruptError("Tag containing meta information is corrupt.");
 	    }
 	    tot_totlen += totlen;
-	    if (tot_totlen < tot_totlen) {
+	    if (tot_totlen < totlen) {
 		throw "totlen wrapped!";
 	    }
-	    if (cur->next()) {
-		pq.push(cur);
-	    } else {
-		delete cur;
-	    }
+	}
+	if (cur->next()) {
+	    pq.push(cur);
 	} else {
-	    delete in;
+	    delete cur;
 	}
     }
 
@@ -410,7 +412,7 @@ merge_spellings(FlintTable * out,
 	pq.pop();
 
 	string key = cur->current_key;
-	if (pq.top()->current_key > key) {
+	if (pq.empty() || pq.top()->current_key > key) {
 	    // No need to merge the tags, just copy the (possibly compressed)
 	    // tag value.
 	    bool compressed = cur->read_tag(true);
@@ -731,6 +733,7 @@ merge_docid_keyed(FlintTable *out, const vector<string> & inputs,
 int
 main(int argc, char **argv)
 {
+    const char * opts = "b:nFm";
     const struct option long_opts[] = {
 	{"fuller",	no_argument, 0, 'F'},
 	{"no-full",	no_argument, 0, 'n'},
@@ -748,7 +751,7 @@ main(int argc, char **argv)
     bool renumber = true;
 
     int c;
-    while ((c = gnu_getopt_long(argc, argv, "b:nFm", long_opts, 0)) != EOF) {
+    while ((c = gnu_getopt_long(argc, argv, opts, long_opts, 0)) != -1) {
 	switch (c) {
 	    case 'b': {
 		char *p;
