@@ -28,13 +28,11 @@
 #include "multimatch.h"
 #include "submatch.h"
 #include "localmatch.h"
-#include "emptysubmatch.h"
 #include "omdebug.h"
 #include "omenquireinternal.h"
 
 #include "emptypostlist.h"
 #include "branchpostlist.h"
-#include "leafpostlist.h"
 #include "mergepostlist.h"
 
 #include "document.h"
@@ -54,7 +52,8 @@
 #endif /* XAPIAN_HAS_REMOTE_BACKEND */
 
 #include <algorithm>
-#include <queue>
+#include <cfloat> // For DBL_EPSILON.
+#include <climits> // For UINT_MAX.
 #include <vector>
 #include <map>
 #include <set>
@@ -100,9 +99,9 @@ split_rset_by_db(const Xapian::RSet * rset,
 		subrsets.push_back(Xapian::RSet());
 	    }
 
-	    const set<Xapian::docid> & items = rset->internal->get_items();
+	    const set<Xapian::docid> & rsetitems = rset->internal->get_items();
 	    set<Xapian::docid>::const_iterator j;
-	    for (j = items.begin(); j != items.end(); ++j) {
+	    for (j = rsetitems.begin(); j != rsetitems.end(); ++j) {
 		Xapian::doccount local_docid = (*j - 1) / number_of_subdbs + 1;
 		Xapian::doccount subdatabase = (*j - 1) % number_of_subdbs;
 		subrsets[subdatabase].add_document(local_docid);
@@ -134,7 +133,7 @@ split_rset_by_db(const Xapian::RSet * rset,
  *  statistics arrive, we can move on to the next step.
  */
 static void
-prepare_sub_matches(std::vector<Xapian::Internal::RefCntPtr<SubMatch> > & leaves,
+prepare_sub_matches(vector<Xapian::Internal::RefCntPtr<SubMatch> > & leaves,
 		    Xapian::ErrorHandler * errorhandler,
 		    Stats & stats)
 {
@@ -150,7 +149,8 @@ prepare_sub_matches(std::vector<Xapian::Internal::RefCntPtr<SubMatch> > & leaves
 	for (size_t leaf = 0; leaf < leaves.size(); ++leaf) {
 	    if (prepared[leaf]) continue;
 	    try {
-		if (leaves[leaf]->prepare_match(nowait, stats)) {
+		SubMatch * submatch = leaves[leaf].get();
+		if (!submatch || submatch->prepare_match(nowait, stats)) {
 		    prepared[leaf] = true;
 		    --unprepared;
 		}
@@ -160,7 +160,7 @@ prepare_sub_matches(std::vector<Xapian::Internal::RefCntPtr<SubMatch> > & leaves
 		DEBUGLINE(EXCEPTION, "Calling error handler for prepare_match() on a SubMatch.");
 		(*errorhandler)(e);
 		// Continue match without this sub-match.
-		leaves[leaf] = new EmptySubMatch();
+		leaves[leaf] = NULL;
 		prepared[leaf] = true;
 		--unprepared;
 	    }
@@ -221,13 +221,13 @@ MultiMatch::MultiMatch(const Xapian::Database &db_,
 #ifdef XAPIAN_HAS_REMOTE_BACKEND
 	    RemoteDatabase *rem_db = subdb->as_remotedatabase();
 	    if (rem_db) {
-		is_remote[i] = true;
 		rem_db->set_query(query, qlen, collapse_key, order, sort_key,
 				  sort_by, sort_value_forward, percent_cutoff,
 				  weight_cutoff, weight, subrsets[i]);
 		bool decreasing_relevance =
 		    (sort_by == REL || sort_by == REL_VAL);
 		smatch = new RemoteSubMatch(rem_db, decreasing_relevance);
+		is_remote[i] = true;
 	    } else {
 #endif /* XAPIAN_HAS_REMOTE_BACKEND */
 		smatch = new LocalSubMatch(subdb, query, qlen, subrsets[i], weight);
@@ -239,7 +239,7 @@ MultiMatch::MultiMatch(const Xapian::Database &db_,
 	    DEBUGLINE(EXCEPTION, "Calling error handler for creation of a SubMatch from a database and query.");
 	    (*errorhandler)(e);
 	    // Continue match without this sub-postlist.
-	    smatch = new EmptySubMatch;
+	    smatch = NULL;
 	}
 	leaves.push_back(smatch);
     }
@@ -301,6 +301,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 
     Assert(!leaves.empty());
 
+#ifdef XAPIAN_HAS_REMOTE_BACKEND
     // If there's only one database and it's remote, we can just unserialise
     // its MSet and return that.
     if (leaves.size() == 1 && is_remote[0]) {
@@ -310,14 +311,14 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	rem_match->get_mset(mset);
 	return;
     }
+#endif
 
     // Start matchers.
     {
 	vector<Xapian::Internal::RefCntPtr<SubMatch> >::iterator leaf;
 	for (leaf = leaves.begin(); leaf != leaves.end(); ++leaf) {
+	    if (!(*leaf).get()) continue;
 	    try {
-		// FIXME: look if we can push the "check_at_least" stuff
-		// into the remote match handling too.
 		(*leaf)->start_match(0, first + maxitems,
 				     first + check_at_least, stats);
 	    } catch (Xapian::Error & e) {
@@ -326,7 +327,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 			  "start_match() on a SubMatch.");
 		(*errorhandler)(e);
 		// Continue match without this sub-match.
-		*leaf = Xapian::Internal::RefCntPtr<SubMatch>(new EmptySubMatch());
+		*leaf = NULL;
 	    }
 	}
     }
@@ -367,7 +368,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	    (*errorhandler)(e);
 	    // FIXME: check if *ALL* the remote servers have failed!
 	    // Continue match without this sub-match.
-	    leaves[i] = new EmptySubMatch();
+	    leaves[i] = NULL;
 	    pl = new EmptyPostList;
 	}
 	postlists.push_back(pl);
@@ -387,10 +388,13 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
     // Empty result set
     Xapian::doccount docs_matched = 0;
     Xapian::weight greatest_wt = 0;
+#ifdef XAPIAN_HAS_REMOTE_BACKEND
+    unsigned greatest_wt_subqs_db_num = UINT_MAX;
+#endif
     vector<Xapian::Internal::MSetItem> items;
 
     // maximum weight a document could possibly have
-    const Xapian::weight max_weight = pl->recalc_maxweight();
+    const Xapian::weight max_possible = pl->recalc_maxweight();
 
     DEBUGLINE(MATCH, "pl = (" << pl->get_description() << ")");
     recalculate_w_max = false;
@@ -421,7 +425,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 					   matches_upper_bound,
 					   matches_lower_bound,
 					   matches_estimated,
-					   max_weight, greatest_wt, items,
+					   max_possible, greatest_wt, items,
 					   termfreqandwts,
 					   0));
 	return;
@@ -457,6 +461,9 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 
     // Factor to multiply maximum weight seen by to get the cutoff weight.
     Xapian::weight percent_cutoff_factor = percent_cutoff / 100.0;
+    // Corresponding correction to that in omenquire.cc to account for excess
+    // precision on x86.
+    percent_cutoff_factor -= DBL_EPSILON;
 
     // Table of keys which have been seen already, for collapsing.
     map<string, pair<Xapian::Internal::MSetItem,Xapian::weight> > collapse_tab;
@@ -481,6 +488,8 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
     bool is_heap = false;
 
     while (true) {
+	bool pushback;
+
 	if (rare(recalculate_w_max)) {
 	    if (min_weight > 0.0) {
 		if (rare(getorrecalc_maxweight(pl) < min_weight)) {
@@ -523,6 +532,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	    calculated_weight = true;
 	}
 
+	Xapian::Internal::RefCntPtr<Xapian::Document::Internal> doc;
 	Xapian::docid did = pl->get_docid();
 	DEBUGLINE(MATCH, "Candidate document id " << did << " wt " << wt);
 	Xapian::Internal::MSetItem new_item(wt, did);
@@ -531,7 +541,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	    Assert(multiplier != 0);
 	    Xapian::doccount n = (new_item.did - 1) % multiplier; // which actual database
 	    Xapian::docid m = (new_item.did - 1) / multiplier + 1; // real docid in that database
-	    Xapian::Internal::RefCntPtr<Xapian::Document::Internal> doc(db.internal[n]->open_document(m, true));
+	    doc = db.internal[n]->open_document(m, true);
 	    if (sorter) {
 		new_item.sort_key = (*sorter)(Xapian::Document(doc.get()));
 	    } else {
@@ -549,6 +559,8 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 		    // processing needed.
 		    DEBUGLINE(MATCH, "Making note of match item which sorts lower than min_item");
 		    ++docs_matched;
+		    if (!calculated_weight) wt = pl->get_weight();
+		    if (wt > greatest_wt) goto new_greatest_weight;
 		    continue;
 		}
 		if (docs_matched >= check_at_least) {
@@ -563,8 +575,6 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	    }
 	}
 
-	Xapian::Internal::RefCntPtr<Xapian::Document::Internal> doc;
-
 	// Use the match spy and/or decision functors (if specified).
 	if (matchspy != NULL || mdecider != NULL) {
 	    const unsigned int multiplier = db.internal.size();
@@ -575,9 +585,8 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	    if (!is_remote[n]) {
 		if (doc.get() == 0) {
 		    Xapian::docid m = (did - 1) / multiplier + 1; // real docid in that database
-
-		    Xapian::Internal::RefCntPtr<Xapian::Document::Internal> temp(db.internal[n]->open_document(m, true));
-		    doc = temp;
+		    doc = db.internal[n]->open_document(m, true);
+		    Assert(doc.get());
 		}
 		Xapian::Document mydoc(doc.get());
 
@@ -599,7 +608,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	    new_item.wt = wt;
 	}
 
-	bool pushback = true;
+	pushback = true;
 	documents_considered++;
 
 	// Perform collapsing on key if requested.
@@ -632,6 +641,11 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 			if (new_item.wt > oldkey->second.second) {
 			    oldkey->second.second = new_item.wt;
 			}
+			// If we're sorting by relevance primarily, then we
+			// throw away the lower weighted document anyway.
+			if (sort_by != REL && sort_by != REL_VAL) {
+			    if (wt > greatest_wt) goto new_greatest_weight;
+			}
 			continue;
 		    }
 		    // Make a note of the updated collapse count in the
@@ -700,7 +714,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 			    // We're done if this is a forward boolean match
 			    // with only one database (bodgetastic, FIXME
 			    // better if we can!)
-			    if (rare(max_weight == 0 && sort_forward)) {
+			    if (rare(max_possible == 0 && sort_forward)) {
 				// In the multi database case, MergePostList
 				// currently processes each database
 				// sequentially (which actually may well be
@@ -728,7 +742,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 			// We're done if this is a forward boolean match
 			// with only one database (bodgetastic, FIXME
 			// better if we can!)
-			if (rare(max_weight == 0 && sort_forward)) {
+			if (rare(max_possible == 0 && sort_forward)) {
 			    // In the multi database case, MergePostList
 			    // currently processes each database
 			    // sequentially (which actually may well be
@@ -743,7 +757,19 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 
 	// Keep a track of the greatest weight we've seen.
 	if (wt > greatest_wt) {
+new_greatest_weight:
 	    greatest_wt = wt;
+#ifdef XAPIAN_HAS_REMOTE_BACKEND
+	    const unsigned int multiplier = db.internal.size();
+	    unsigned int db_num = (did - 1) % multiplier;
+	    if (is_remote[db_num]) {
+		// Note that the greatest weighted document came from a remote
+		// database, and which one.
+		greatest_wt_subqs_db_num = db_num;
+	    } else {
+		greatest_wt_subqs_db_num = UINT_MAX;
+	    }
+#endif
 	    if (percent_cutoff) {
 		Xapian::weight w = wt * percent_cutoff_factor;
 		if (w > min_weight) {
@@ -780,21 +806,26 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	vector<Xapian::Internal::MSetItem>::const_iterator best;
 	best = min_element(items.begin(), items.end(), mcmp);
 
-	unsigned int multiplier = db.internal.size();
-	Assert(multiplier != 0);
-	Xapian::doccount n = (best->did - 1) % multiplier; // which actual database
-	// If the top result is from a remote database, then we can
-	// just use the percentage scaling calculated for that
-	// database.
-	if (is_remote[n]) {
+#ifdef XAPIAN_HAS_REMOTE_BACKEND
+	if (greatest_wt_subqs_db_num != UINT_MAX) {
+	    const unsigned int n = greatest_wt_subqs_db_num;
 	    RemoteSubMatch * rem_match;
 	    rem_match = static_cast<RemoteSubMatch*>(leaves[n].get());
-	    percent_scale = rem_match->get_percent_factor();
-	} else if (termfreqandwts.size() > 1) {
+	    percent_scale = rem_match->get_percent_factor() / 100.0;
+	} else
+#endif
+	if (termfreqandwts.size() > 1) {
 	    Xapian::termcount matching_terms = 0;
 	    map<string,
 		Xapian::MSet::Internal::TermFreqAndWeight>::const_iterator i;
 
+	    // Special case for MatchAll queries.
+	    i = termfreqandwts.find(string());
+	    if (i != termfreqandwts.end()) {
+		percent_scale += i->second.termweight;
+		++matching_terms;
+	    }
+
 	    Xapian::TermIterator docterms = db.termlist_begin(best->did);
 	    Xapian::TermIterator docterms_end = db.termlist_end(best->did);
 	    while (docterms != docterms_end) {
@@ -806,12 +837,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 		}
 		++docterms;
 	    }
-	    // Special case for MatchAll queries
-	    i = termfreqandwts.find("");
-	    if (i != termfreqandwts.end()) {
-		percent_scale += i->second.termweight;
-		++matching_terms;
-	    }
+
 	    if (matching_terms < termfreqandwts.size()) {
 		// OK, work out weight corresponding to 100%
 		double denom = 0;
@@ -834,8 +860,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	}
 	Assert(percent_scale > 0);
 	if (percent_cutoff) {
-	    // FIXME: better to sort and binary chop maybe?  we
-	    // could use the sort above to find "best" too.
+	    // FIXME: better to sort and binary chop maybe?
 	    // Or we could just use a linear scan here instead.
 
 	    // trim the mset to the correct answer...
@@ -888,8 +913,8 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	matches_lower_bound = matches_upper_bound = matches_estimated
 	    = docs_matched;
     } else {
-	Assert(matches_estimated >= matches_lower_bound);
-	Assert(matches_estimated <= matches_upper_bound);
+	AssertRel(matches_estimated,>=,matches_lower_bound);
+	AssertRel(matches_estimated,<=,matches_upper_bound);
 
 	// We can end up scaling the estimate more than once, so collect
 	// the scale factors and apply them in one go to avoid rounding
@@ -978,11 +1003,11 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 		      "matches_lower_bound = " << matches_lower_bound <<
 		      ", matches_estimated = " << matches_estimated <<
 		      ", matches_upper_bound = " << matches_upper_bound);
-	    Assert(matches_lower_bound <= matches_upper_bound);
+	    AssertRel(matches_lower_bound,<=,matches_upper_bound);
 	    matches_estimated = max(matches_estimated, matches_lower_bound);
 	    matches_estimated = min(matches_estimated, matches_upper_bound);
 	} else if (!percent_cutoff) {
-	    Assert(docs_matched <= matches_upper_bound);
+	    AssertRel(docs_matched,<=,matches_upper_bound);
 	    if (docs_matched > matches_lower_bound)
 		matches_lower_bound = docs_matched;
 	    if (docs_matched > matches_estimated)
@@ -1020,8 +1045,8 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	DEBUGLINE(MATCH, "max weight in mset = " << items[0].wt);
     }
 
-    Assert(matches_estimated >= matches_lower_bound);
-    Assert(matches_estimated <= matches_upper_bound);
+    AssertRel(matches_estimated,>=,matches_lower_bound);
+    AssertRel(matches_estimated,<=,matches_upper_bound);
 
     // We may need to qualify any collapse_count to see if the highest weight
     // collapsed item would have qualified percent_cutoff
@@ -1066,7 +1091,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 				       matches_upper_bound,
 				       matches_lower_bound,
 				       matches_estimated,
-				       max_weight, greatest_wt, items,
+				       max_possible, greatest_wt, items,
 				       termfreqandwts,
 				       percent_scale));
 }
